"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean advection'

# --------------------------------------------------------------------
# SUB-PROCESS: Momentum.
# --------------------------------------------------------------------
DETAILS['momentum'] = {
    'description': 'Properties of lateral momemtum advection scheme in ocean',
    'properties': [
        ('type', 'ENUM:adv_mom_scheme_types', '1.1',
            'Type of lateral momemtum advection scheme in ocean'),
        ('scheme_name', 'str', '1.1',
            'Name of ocean momemtum advection scheme'),
        ('ALE', 'bool', '0.1',
            'Using ALE for vertical advection ? (if vertical coordinates are sigma)'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Lateral tracers.
# --------------------------------------------------------------------
DETAILS['lateral_tracers'] = {
    'description':'Properties of lateral tracer advection scheme in ocean',
    'properties': [
        ('type', 'ENUM:adv_tra_scheme_types', '1.1',
            "Type of lateral tracer advection scheme in ocean"),
        ('flux_limiter', 'bool', '1.1',
            "Monotonic flux limiter for vertical tracer advection scheme in ocean ?"),
        ('passive_tracers','ENUM:passive_tracers_list','0.N',
            "Passive tracers advected"),
        ('passive_tracers_advection','str','0.1',
            "Is advection of passive tracers different than active ? if so, describe.")
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Vertical tracers.
# --------------------------------------------------------------------
DETAILS['vertical_tracers'] = {
    'description': 'Properties of vertical tracer advection scheme in ocean',
    'properties': [
        ('type', 'ENUM:adv_tra_scheme_types', '1.1',
            "Type of vertical tracer advection scheme in ocean"),
        ('flux_limiter', 'bool', '1.1',
            "Monotonic flux limiter for vertical tracer advection scheme in ocean ?"),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['adv_mom_scheme_types'] = {
    'description': 'Type of lateral momemtum advection scheme in ocean',
    'is_open': False,
    'members': [
        ('Flux form', None),
        ('Vector form', None)
        ]
    }

ENUMERATIONS['adv_tra_scheme_types'] = {
    'description': 'Type of tracer advection scheme in ocean',
    'is_open': True,
    'members': [
        ('Centred 2nd order', None),
        ('Centred 4th order', None),
        ('Total Variance Dissipation (TVD)', None),
        ('MUSCL', None),
        ('QUICKEST', None),
        ('Piecewise Parabolic method', None),
        ('Sweby', None),
        ('Prather 2nd moment (PSOM)', None),
        ('3rd order upwind', None)
        ]
    }

ENUMERATIONS['passive_tracers_list'] = {
    'description': 'Passive tracers in ocean',
    'is_open': True,
    'members': [
        ('Ideal age', None),
        ('CFC 11', None),
        ('CFC 12', None),
        ('SF6', None)
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.class
   :platform: Unix, Windows
   :synopsis: Represents an ontological class definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import itertools
from collections import defaultdict

import constants


class Class(object):
    """Represents an ontological class definition.

    """
    def __init__(
        self,
        name,
        base,
        is_abstract,
        is_document,
        doc_string,
        properties,
        computed_properties,
        constraints,
        decodings,
        pstr = None
        ):
        """Instance constructor.

        :param str name: Class name.
        :param str base: Base class used in object hierarchies.
        :param bool is_abstract: Flag indicating whether this is an abstract class or not.
        :param bool is_document: Flag indicating whether this is a document or not.
        :param str doc_string: Class documentation string.
        :param list properties: Set of associated properties.
        :param list computed_properties: Set of associated computed properties.
        :param list constraints: Set of associated constraints.
        :param list decodings: Set of associated property decodings.
        :param ClassPrintString pstr: Class print string information.

        """
        self.base = base
        self.circular_imports = []
        self.cls = None
        self.computed_properties = sorted(computed_properties, key=lambda p: p.name)
        self.constraints = sorted(constraints, key=lambda ct: ct.property_name)
        self.decodings = sorted(decodings, key=lambda dc: dc.property_name)
        self.doc_string = doc_string if doc_string is not None else ''
        self.imports = []
        self.is_abstract = is_abstract
        self.is_document = is_document
        self.is_document_meta = name.endswith("doc_meta_info")
        self.name = name
        self.sub_class_hierachy = tuple()
        self.sub_classes = tuple()
        self.pstr = pstr
        self.op_base_name = None
        self.op_doc_string_name = None
        self.op_file_name = None
        self.op_full_name = None
        self.op_func_name = None
        self.op_import_name = None
        self.op_name = None
        self.properties = sorted(properties, key=lambda p: p.name)
        self.package = None


    def __repr__(self):
        """Instance string representation.

        """
        try:
            return "{}.{}".format(self.package, self.name)
        except:
            return self.name


    @property
    def is_entity(self):
        """Gets a flag indicating whether this class is considered an entity.

        """
        if self.is_document == True:
            return True
        if self.base:
            return self.base.is_entity
        return False


    @property
    def sub_class_hierachy_depth(self):
        """Gets depth of sub-class hierachy.

        """
        return len(self.sub_class_hierachy)


    @property
    def constants(self):
        """Gets collection of constant constraints.

        """
        return [(ct.property_name, ct.value) for ct in self.constraints
                if ct.typeof == constants.CONSTRAINT_TYPE_CONSTANT]


    @property
    def bases(self):
        """Gets collection of constant constraints.

        """
        if self.base:
            return (self.base, ) + self.base.bases
        return tuple()


    @property
    def all_constraints(self):
        """Gets all associated constraints including those of base class (sorted by name).

        Note that child constraints overrides parent constraints.

        """
        result = defaultdict(lambda: defaultdict(list))

        # Own constraints.
        for ct in self.constraints:
            if ct.property_name not in result[ct.typeof]:
                result[ct.typeof][ct.property_name] = (ct.property_name, ct.typeof, ct.value)

        # Base class(es) constraints.
        if self.base:
            for n, t, v in self.base.all_constraints:
                if n not in result[t]:
                    result[t][n] = (n, t, v)

        # Own properties converted to constraints.
        for p in self.all_properties:
            if p.name not in result['cardinality']:
                result['cardinality'][p.name] = (p.name, "cardinality", p.cardinality)
            if p.name not in result['type']:
                result['type'][p.name] = (p.name, "type", p.type)

        return list(itertools.chain.from_iterable([v.values() for v in result.values()]))


    @property
    def all_properties(self):
        """Gets all associated properties including those of base class (sorted by name).

        """
        if self.base:
            return self.properties + self.base.all_properties
        return self.properties


    @property
    def inherited_properties(self):
        """Gets properties inherited from base classes.

        """
        if self.base:
            return self.base.all_properties
        return []


    @property
    def all_computed_properties(self):
        """Gets all associated computed properties including those of base class (sorted by name).

        """
        if self.base:
            return self.computed_properties + self.base.all_computed_properties
        return self.computed_properties


    @property
    def all_decodings(self):
        """Gets class plus base class decodings.

        """
        if self.base:
            return self.decodings + self.base.all_decodings
        return self.decodings


    def get_property_decodings(self, prp):
        """Returns set of property decodings.

        :param pyesdoc.mp.ontologies.core.Property prp: A property being processed.

        """
        return [dc for dc in self.all_decodings if dc.property_name == prp.name]


    def get_property(self, name):
        """Recursively gets associated property either from self or from base class.

        :param str name: Name of a property.

        """
        for prp in self.properties:
            if prp.name == name:
                return prp
        if self.base:
            return self.base.get_property(name)




# -*- coding: utf-8 -*-

"""
.. module:: model.py
   :platform: Unix, Windows
   :synopsis: A repesentation of CMIP6 specializations.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from itertools import chain

from utils_constants import *



class TopicSpecialization(object):
    """Wraps a topic specialization.

    """
    def __init__(self, spec, type_key):
        """Instance initializer.

        """
        self.authors = None
        self.contact = None
        self.contributors = None
        self.description = None
        self.id = None
        self.qc_status = None
        self.name = None
        self.parent = None
        self.properties = []
        self.property_sets = []
        self.spec = spec
        self.sub_topics = []
        self.type_key = type_key


    def __repr__(self):
        """Instance representation.

        """
        return self.id


    def __getitem__(self, type_key):
        """Returns a child topic.

        """
        result = [i for i in self.sub_topics if i.type_key == type_key]

        if type_key in {TYPE_KEY_PROCESS, TYPE_KEY_SUBPROCESS}:
            return result
        elif len(result) > 1:
            return result
        elif len(result) == 1:
            return result[0]


    @property
    def name_camel_case(self):
        """Gets camel case formatted name.

        """
        return _to_camel_case(self.name)


    @property
    def name_camel_case_spaced(self):
        """Gets spaced camel case formatted name.

        """
        return _to_camel_case_spaced(self.name)


    @property
    def all_topics(self):
        """Returns a flattened topic hierarchy.

        """
        def _get(container):
            result = [container]
            for topic in container.sub_topics:
                result += _get(topic)

            return result

        return _get(self)


    @property
    def all_property_containers(self):
        """Returns flattened list of all property containers, i.e. topics & proeprty-sets.

        """
        result = []
        for t in self.all_topics:
            result += [t] + t.property_sets

        return [i for i in result if i.properties]


    @property
    def all_sub_topics(self):
        """Returns a flattened topic hierarchy.

        """
        return self.all_topics[1:]


    @property
    def all_properties(self):
        """Returns all specialization properties.

        """
        return set(chain.from_iterable(i.properties for i in self.all_property_containers))


    @property
    def all_required_properties(self):
        """Returns all required specialization properties.

        """
        return set([i for i in self.all_properties if i.is_required])


    @property
    def all_optional_properties(self):
        """Returns all optional specialization properties.

        """
        return self.all_properties - self.all_required_properties


    @property
    def ENUMS(self):
        """Gets associated enumeration definitions.

        """
        try:
            return self.spec.ENUMERATIONS
        except AttributeError:
            try:
                return self.parent.spec.ENUMERATIONS
            except AttributeError:
                return []


    def names(self, offset=None, seperator=" --> ", convertor=None):
        """Returns set of topic names derived from topic specialization id.

        """
        if convertor is None:
            convertor = _to_camel_case_spaced
        names = self.id.split(".")
        if offset is not None:
            names = names[offset:]
        names = [convertor(i) for i in names]

        return seperator.join(names)


    def has_property(self, property_id):
        """Returns a flag indicating whether a topic supports a property.

        """
        for prop in self.all_properties:
            if prop.id == property_id:
                return True
        return False


class PropertySetSpecialization(object):
    """Wraps a property set specialization.

    """
    def __init__(self):
        """Instance initializer.

        """
        super(PropertySetSpecialization, self).__init__()

        self.description = None
        self.id = None
        self.name = None
        self.owner = None
        self.properties = []
        self.property_sets = []
        self.topic = None
        self.type_key = "property-set"


    def __repr__(self):
        """Instance representation.

        """
        return self.id


    @property
    def name_camel_case(self):
        """Gets camel case formatted name.

        """
        return _to_camel_case(self.name)


    @property
    def name_camel_case_spaced(self):
        """Gets spaced camel case formatted name.

        """
        return _to_camel_case_spaced(self.name)


    def names(self, offset=None, seperator=" --> ", convertor=None):
        """Returns set of topic names derived from topic specialization id.

        """
        if convertor is None:
            convertor = _to_camel_case_spaced
        names = self.id.split(".")
        if offset is not None:
            names = names[offset:]
        names = [convertor(i) for i in names]
        return seperator.join(names)


class PropertySpecialization(object):
    """Wraps a property specialization.

    """
    def __init__(self):
        """Instance initializer.

        """
        super(PropertySpecialization, self).__init__()

        self.cardinality = None
        self.description = None
        self.enum = None
        self.id = None
        self.name = None
        self.owner = None
        self.topic = None
        self.typeof = None
        self.type_key = "property"


    def __repr__(self):
        """Instance representation.

        """
        return self.id


    @property
    def name_camel_case(self):
        """Gets camel case formatted name.

        """
        return _to_camel_case(self.name)


    @property
    def name_camel_case_spaced(self):
        """Gets spaced camel case formatted name.

        """
        return _to_camel_case_spaced(self.name)


    @property
    def is_required(self):
        """Gets flag indicating whether cardinality is mandatory or not.

        """
        return self.cardinality.split(".")[0] != "0"


    @property
    def is_collection(self):
        """Gets flag indicating whether property is a collection or not.

        """
        try:
            int(self.cardinality.split(".")[1])
        except ValueError:
            return True
        return False


    @property
    def typeof_label(self):
        """Gets label for the property type.

        """
        if self.typeof == 'str':
            return "STRING"
        elif self.typeof == 'bool':
            return "BOOLEAN"
        elif self.typeof == 'int':
            return "INTEGER"
        elif self.typeof == 'float':
            return "FLOAT"
        elif self.enum:
            return "ENUM"
        return self.typeof.split(":")[-1].upper()


    def validate_value(self, val):
        """Validates a property value.

        :param object val: Value to be validated.

        """
        if self.enum:
            self.enum.validate_value(val)

        if self.typeof == 'str':
            if not isinstance(val, basestring) or not len(val.strip()):
                raise ValueError("Invalid value: must be a non zero length string")

        if self.typeof == 'bool':
            if not isinstance(val, bool):
                raise ValueError("Invalid value: must be a boolean")

        if self.typeof == 'int':
            if not isinstance(val, int):
                raise ValueError("Invalid value: must be an integer")

        if self.typeof == 'float':
            if not isinstance(val, float):
                raise ValueError("Invalid value: must be a float")


class EnumSpecialization(object):
    """Wraps an enumeration specialization.

    """
    def __init__(self):
        """Instance initializer.

        """
        self.choices = []
        self.description = None
        self.id = None
        self.is_open = False
        self.label = None
        self.name = None
        self.type_key = "enum"


    def __repr__(self):
        """Instance representation.

        """
        return self.id


    def is_a_member(self, val):
        """Returns flag indicating whether vlue is a member of the enumeration.

        :param object val: Value to be validated.

        """
        return val in [i.value for i in self.choices]


    def validate_value(self, val):
        """Validates a value against the set of enum choices.

        :param object val: Value to be validated.

        """
        if not isinstance(val, basestring) or not len(val.strip()):
            raise ValueError("Invalid value: must be a non zero length string")
        if self.is_a_member(val):
            return
        if not self.is_open:
            raise ValueError("Invalid value: is not an enumeration member")
        if not val.startswith("Other: "):
            raise ValueError("Invalid value: new enumeration members must be prefixed with: 'Other: '")
        other = val.split("Other: ")[-1]
        if self.is_a_member(other):
            raise ValueError("Invalid value: enumeration member already defined")


class EnumChoiceSpecialization(object):
    """Wraps an enumeration choice specialization.

    """
    def __init__(self):
        """Instance initializer.

        """
        self.description = None
        self.enum = None
        self.id = None
        self.value = None
        self.type_key = "enum-choice"

        self.is_other = False


    def __repr__(self):
        """Instance representation.

        """
        return self.id


def _to_camel_case_spaced(name, separator='_'):
    """Converts passed name to camel case with space.

    :param str name: A name as specified in ontology specification.
    :param str separator: Separator to use in order to split name into constituent parts.

    """
    s = _to_camel_case(name, separator)
    r = s[0]
    for s in s[1:]:
        if s.upper() == s:
            r += " "
        r += s

    return r


def _to_camel_case(name, separator='_'):
    """Converts passed name to camel case.

    :param str name: A name as specified in ontology specification.
    :param str separator: Separator to use in order to split name into constituent parts.

    """
    r = ''
    if name is not None:
        s = name.split(separator)
        for s in s:
            if (len(s) > 0):
                r += s[0].upper()
                if (len(s) > 1):
                    r += s[1:]
    return r


# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.enum_member
   :platform: Unix, Windows
   :synopsis: Represents an ontological enumerated type member definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

class EnumMember(object):
    """Represents an enumeration member, i.e. a constrained value associated with an enumeration.

    """
    def __init__(self, name, doc_string):
        """Instance constructor.

        :param str name: Enumeration member name.
        :param str doc_string: Enumeration member documentation string.

        """
        self.enum = None
        self.name = name
        self.doc_string = doc_string


    def __repr__(self):
        """Instance string representation.

        """
        return self.name


    @property
    def key(self):
        """Returns enum member key.

        """
        return self.name.strip().lower().replace(" ", "_")



# -*- coding: utf-8 -*-

"""
.. module:: software_model_component_1.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.software.modelcomponent document.


"""
from collections import namedtuple

from pyesdoc.utils.convert import capitalize
from pyesdoc.utils.convert import str_to_spaced_case
from pyesdoc.utils.convert import str_to_unicode




# Property extension context information.
_ExtensionContextInfo = \
    namedtuple('_ExtensionContextInfo', ['c', 'p', 'parent', 'ancestors'])


class _PropertyExtensionInfo(object):
    """Property extension properties.

    """
    def __init__(self):
        self.display_name = unicode()
        self.full_display_name = unicode()
        self.long_display_name = unicode()
        self.short_display_name = unicode()


def _extend_property_01(ctx):
    """Sets property extension properties.

    """
    ctx.p.ext = _PropertyExtensionInfo()
    ctx.p.ext.ancestors = ctx.ancestors
    ctx.p.ext.depth = len(ctx.ancestors)
    ctx.p.ext.parent = ctx.parent


def _extend_property_02(ctx):
    """Formats property values.

    """
    ctx.p.values = filter(None, ctx.p.values)
    ctx.p.values = map(str_to_unicode, ctx.p.values)
    ctx.p.values = filter(len, ctx.p.values)
    ctx.p.values = map(lambda v: v[0].upper() + v[1:], ctx.p.values)


def _extend_property_03(ctx):
    """Sets property display names.

    """
    # Short name.
    name = ctx.p.short_name
    if len(name) and not name[0].isdigit():
        name = str_to_spaced_case(name)
        name = map(capitalize, name.split(" "))
        name = reduce(lambda x, s: s if x is None else x + " " + s, name, None)
    ctx.p.ext.short_display_name = name

    # Display name.
    if ctx.p.ext.depth:
        ctx.p.ext.display_name += ctx.p.ext.parent.ext.display_name
        ctx.p.ext.display_name += u" > "
    ctx.p.ext.display_name += ctx.p.ext.short_display_name

    # Long name.
    if not ctx.p.ext.depth:
        ctx.p.ext.long_display_name = ctx.c.ext.long_display_name
        ctx.p.ext.long_display_name += u" >> "
    else:
        ctx.p.ext.long_display_name = ctx.p.ext.parent.ext.long_display_name
        ctx.p.ext.long_display_name += u" > "
    ctx.p.ext.long_display_name += ctx.p.ext.short_display_name

    # Full name.
    if not ctx.p.ext.depth:
        ctx.p.ext.full_display_name = ctx.c.ext.full_display_name
        ctx.p.ext.full_display_name += u" >> "
    else:
        ctx.p.ext.full_display_name = ctx.p.ext.parent.ext.full_display_name
        ctx.p.ext.full_display_name += u" > "
    ctx.p.ext.full_display_name += ctx.p.ext.short_display_name


def _extend_property_04(ctx):
    """Process child properties.

    """
    for p in ctx.p.sub_properties:
        extend(ctx.c, p, ctx.p, ctx.ancestors + [ctx.p])


def extend(c, p, parent=None, ancestors=[]):
    """Extends a property in readiness for later processing.

    :param c: A model component.
    :type c: pyesdoc.ontologies.cim.v1.software.ModelComponent

    :param p: A model component property.
    :type p: pyesdoc.ontologies.cim.v1.software.ComponentProperty

    :param parent: Parent model component property.
    :type parent: pyesdoc.ontologies.cim.v1.software.ComponentProperty

    :param list ancestors: Ancestor model component properties.

    """
    ctx = _ExtensionContextInfo(c, p, parent, ancestors)
    for func in (
        _extend_property_01,
        _extend_property_02,
        _extend_property_03,
        _extend_property_04,
        ):
        try:
            func(ctx)
        except Exception as err:
            print func, err
            raise err



# Import pyesdoc
# ... pip install pyesdoc before running notebook
import pyesdoc

# Initialise data request.
# ... loads, parses & maps dreq.xml & dreqDefn.xml
pyesdoc.drq.initialize()

# Set helper pointers.
DRQ_DEFN = pyesdoc.drq.definition
DRQ = pyesdoc.drq.content

for table in DRQ_DEFN:
    for attribute in table:
        pass

# Iterate section item properties.
# for section in DRQ:
#     for item in section:
#     	print item.links.__dict__
#         # print section, item
#         for attr, value in item:
#             pass

# for section in DRQ:
#     for item in section:
#         for attr, value in item:
#             if not isinstance(value, (attr.type_python, type(None))):
#                 print section, item, attr, value, attr.type_python



"""A model key-properties sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to top-level specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to top-level specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Key properties of the model'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel:flux_correction'] = {
    'description': 'Flux correction properties of the model',
    'properties':[
        ('details', 'str', '1.1',
            'Describe if/how flux corrections are applied in the model'),
        ]
    }

DETAILS['toplevel:genealogy'] = {
    'description': 'Genealogy and history of the model',
    'properties':[
        ('year_released', 'str', '1.1',
            'Year the model was released'),
        ('CMIP3_parent', 'str', '0.1',
            'CMIP3 parent if any'),
        ('CMIP5_parent', 'str', '0.1',
            'CMIP5 parent if any'),
        ('previous_name', 'str', '0.1',
            'Previously known as'),
        ]
    }

DETAILS['toplevel:software_properties'] = {
    'description': 'Software properties of model',
    'properties':[
        ('repository','str', '0.1',
            "Location of code for this component."),
        ('code_version','str', '0.1',
            "Code version identifier."),
        ('code_languages','str', '0.N',
            "Code language(s)."),
        ('components_structure','str', '0.1',
            "Describe how model realms are structured into independent software components (coupled via a coupler) and internal software components."),
        ('coupler','ENUM:coupler_framework', '0.1',
            "Overarching coupling framework for model."),
        ]
    }

# --------------------------------------------------------------------
# TUNING APPLIED: Any tuning used to optimise the parameters in this realm
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': 'Tuning methodology for model',
    'properties': [
        ('description', 'str', '1.1',
            "General overview description of tuning: explain and motivate the main targets and metrics retained. Document the relative weight given to climate performance metrics versus process oriented metrics, and on the possible conflicts with parameterization level tuning. In particular describe any struggle with a parameter value that required pushing it to its limits to solve a particular model deficiency."),
        ('global_mean_metrics_used', 'str', '0.N',
            "List set of metrics of the global mean state used in tuning model"),
        ('regional_metrics_used', 'str', '0.N',
            "List of regional metrics of mean state (e.g THC, AABW, regional means etc) used in tuning model/component"),
        ('trend_metrics_used', 'str', '0.N',
            "List observed trend metrics used in tuning model/component (such as 20th century)"),
        ('energy_balance','str', '1.1',
            "Describe how energy balance was obtained in the full system: in the various components independently or at the components coupling stage?"),
        ('fresh_water_balance','str', '1.1',
            "Describe how fresh_water balance was obtained in the full system: in the various components independently or at the components coupling stage?")
        ]
    }


# --------------------------------------------------------------------
# CONSERVATION PROPERTIES: Global conservation properties of the model.
# --------------------------------------------------------------------
DETAILS['conservation'] = {
    'description': 'Global convervation properties of the model',
    'detail_sets':[
        'heat',
        'fresh_water',
        'salt',
        'momentum'
        ]
    }

DETAILS['conservation:heat'] = {
    'description':'Global heat convervation properties of the model',
    'properties': [
        ('global', 'str', '1.1',
            'Describe if/how heat is conserved globally'),
        ('atmos_ocean_interface', 'str', '0.1',
            'Describe if/how heat is conserved at the atmosphere/ocean coupling interface'),
        ('atmos_land_interface', 'str', '1.1',
            'Describe if/how heat is conserved at the atmosphere/land coupling interface'),
        ('atmos_sea-ice_interface', 'str', '0.1',
            'Describe if/how heat is conserved at the atmosphere/sea-ice coupling interface'),
        ('ocean_seaice_interface', 'str', '0.1',
            'Describe if/how heat is conserved at the ocean/sea-ice coupling interface'),
        ('land_ocean_interface', 'str', '0.1',
            'Describe if/how heat is conserved at the land/ocean coupling interface'),
        ]
    }

DETAILS['conservation:fresh_water'] = {
    'description':'Global fresh water convervation properties of the model',
    'properties': [
        ('global', 'str', '1.1',
            'Describe if/how fresh_water is conserved globally'),
        ('atmos_ocean_interface', 'str', '0.1',
            'Describe if/how fresh_water is conserved at the atmosphere/ocean coupling interface'),
        ('atmos_land_interface', 'str', '1.1',
            'Describe if/how fresh water is conserved at the atmosphere/land coupling interface'),
        ('atmos_sea-ice_interface', 'str', '0.1',
            'Describe if/how fresh water is conserved at the atmosphere/sea-ice coupling interface'),
        ('ocean_seaice_interface', 'str', '0.1',
            'Describe if/how fresh water is conserved at the ocean/sea-ice coupling interface'),
        ('runoff', 'str', '0.1',
            'Describe how runoff is distributed and conserved'),
        ('iceberg_calving', 'str', '0.1',
            'Describe if/how iceberg calving is modeled and conserved'),
        ('endoreic_basins', 'str', '0.1',
            'Describe if/how endoreic basins (no ocean access) are treated'),
        ('snow_accumulation', 'str', '0.1',
            'Describe how snow accumulation over land and over sea-ice is treated'),
        ]
    }

DETAILS['conservation:salt'] = {
    'description':'Global salt convervation properties of the model',
    'properties': [
        ('ocean_seaice_interface', 'str', '0.1',
            'Describe if/how salt is conserved at the ocean/sea-ice coupling interface'),
        ]
    }

DETAILS['conservation:momentum'] = {
    'description':'Global momentum convervation properties of the model',
    'properties': [
        ('details', 'str', '0.1',
            'Describe if/how momentum is conserved in the model'),
        ]
    }

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['coupler_framework'] = {
    'description': 'Coupling software framework.',
    'is_open': True,
    'members': [
        ("OASIS", "The OASIS coupler - prior to OASIS-MCT"),
        ("OASIS3-MCT", "The MCT variant of the OASIS coupler"),
        ("ESMF", "Vanilla Earth System Modelling Framework"),
        ("NUOPC", "National Unified Operational Prediction Capability variant of ESMF"),
        ("Bespoke", "Customised coupler developed for this model"),
        ("Unknown", "It is not known what/if-a coupler is used"),
        ("None", "No coupler is used")
        ]
    }




"""A model radiative forcings sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to top-level specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to top-level specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell, Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Radiative forcings of the model for historical and scenario (aka Table 12.1 IPCC AR5)'

# --------------------------------------------------------------------
# GREENHOUSE GASES
# --------------------------------------------------------------------
DETAILS['greenhouse_gases'] = {
    'description': 'Greenhouse gas forcing agents'
}

DETAILS['greenhouse_gases:CO2'] = {
    'description': 'Carbon dioxide forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['greenhouse_gases:CH4'] = {
    'description': 'Methane forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['greenhouse_gases:N2O'] = {
    'description': 'Nitrous oxide forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['greenhouse_gases:tropospheric_O3'] = {
    'description': 'Troposheric ozone forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['greenhouse_gases:stratospheric_O3'] = {
    'description': 'Stratospheric ozone forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['greenhouse_gases:CFC'] = {
    'description': 'Ozone-depleting and non-ozone-depleting fluorinated gases forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('equivalence_concentration','ENUM:cfc_equivalence_concentration', '1.1',
            'Details of any equivalence concentrations used'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

# --------------------------------------------------------------------
# AEROSOLS
# --------------------------------------------------------------------
DETAILS['aerosols'] = {
    'description': 'Aerosol forcing agents'
}

DETAILS['aerosols:SO4'] = {
    'description': 'SO4 aerosol forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['aerosols:black_carbon'] = {
    'description': 'Black carbon aerosol forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['aerosols:organic_carbon'] = {
    'description': 'Organic carbon aerosol forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['aerosols:nitrate'] = {
    'description': 'Nitrate forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['aerosols:cloud_albedo_effect'] = {
    'description': 'Cloud albedo effect forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('aerosol_effect_on_ice_clouds', 'bool', '1.1',
             'Radiative effects of aerosols on ice clouds are represented?'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).')
    ]
}

DETAILS['aerosols:cloud_lifetime_effect'] = {
    'description': 'Cloud lifetime effect forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('aerosol_effect_on_ice_clouds', 'bool', '1.1',
             'Radiative effects of aerosols on ice clouds are represented?'),
        ('RFaci_from_sulfate_only', 'bool', '1.1',
             'Radiative forcing from aerosol cloud interactions from sulfate aerosol only?'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['aerosols:dust'] = {
    'description': 'Dust forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['aerosols:tropospheric_volcanic'] = {
    'description': 'Tropospheric volcanic forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('historical_explosive_volcanic_aerosol_implementation','ENUM:explosive_volcanic_aerosol_implementation', '1.1',
            'How explosive volcanic aerosol is implemented in historical simulations'),
        ('future_explosive_volcanic_aerosol_implementation','ENUM:explosive_volcanic_aerosol_implementation', '1.1',
            'How explosive volcanic aerosol is implemented in future simulations'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['aerosols:stratospheric_volcanic'] = {
    'description': 'Stratospheric volcanic forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('historical_explosive_volcanic_aerosol_implementation','ENUM:explosive_volcanic_aerosol_implementation', '1.1',
            'How explosive volcanic aerosol is implemented in historical simulations'),
        ('future_explosive_volcanic_aerosol_implementation','ENUM:explosive_volcanic_aerosol_implementation', '1.1',
            'How explosive volcanic aerosol is implemented in future simulations'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['aerosols:sea_salt'] = {
    'description': 'Sea salt forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('additional_information', 'str', '0.1',
            'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

# --------------------------------------------------------------------
# OTHER
# --------------------------------------------------------------------
DETAILS['other'] = {
    'description': 'Miscellaneous forcing agents'
}

DETAILS['other:land_use'] = {
    'description': 'Land use forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('crop_change_only', 'bool', '1.1',
             'Land use change represented via crop change only?'),
        ('additional_information', 'str', '0.1',
             'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

DETAILS['other:solar'] = {
    'description': 'Solar forcing',
    'properties': [
        ('provision', 'ENUM:forcing_provision', '1.N',
             'How this forcing agent is provided (e.g. via concentrations, emission precursors, prognostically derived, etc.)'),
        ('historical_orbital_parameters', 'bool', '1.1',
             'In historical period only, use of realistic time-varying orbital parameters for solar forcing (GISS-E2)'),
        ('additional_information', 'str', '0.1',
             'Additional information relating to the provision and implementation of this forcing agent (e.g. citations, use of non-standard datasets, explaining how multiple provisions are used, etc.).'),
    ]
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['forcing_provision'] = {
    'description': 'How a radiative forcing agent is provided',
    'is_open': True,
    'members': [
        ("N/A", "Not applicable - forcing agent is not included"),
        ("M",  "Emissions and concentrations determined by the model state rather than externally prescribed"),
        ("Y",  "Prescribed concentrations, distributions or time series data"),
        ("E",  "Concentrations calculated interactively driven by prescribed emissions or precursor emissions"),
        ("ES", "Surface emissions (and 3-D concentrations away from the surface) derived via the model from the prescribed surface concentration"),
        ("C",  "Fixed prescribed climatology of concentrations with no year-to-year variability"),
    ]
}

ENUMERATIONS['cfc_equivalence_concentration'] = {
    'description': 'Details of any externally provided equivalence concentrations used for CFC gases',
    'description': '',
    'is_open': True,
    'members': [
        ("N/A", "Not applicabale (CFCs not included or emissions and concentrations determined by the model state)"),
        ("Option 1", "CFCs, including CFC-12, are provided as actual concentrations"),
        ("Option 2", "CFC-12 is provided as actual concentrations and any other gases are provided as an equivalence concentration of CFC-11"),
        ("Option 3", "Ozone depleting gases, including CFC-12, are provided as an equivalence concentration of CFC-12 and all other fluorinated gases are provided as an equivalence concentration of HFC-134a"),
    ]
}

ENUMERATIONS['explosive_volcanic_aerosol_implementation'] = {
    'description': 'Model implementation of explosive volcanic aerosols',
    'is_open': True,
    'members': [
        ("Type A", "Explosive volcanic aerosol returns rapidly to zero (or near-zero) background."),
        ("Type B", "Explosive volcanic aerosol returns rapidly to constant (average volcano)"),
        ("Type C", "Explosive volcanic aerosol returns slowly (over several decades) to constant (average volcano) background."),
        ("Type D", "Explosive volcanic aerosol set to zero"),
        ("Type E", "Explosive volcanic aerosol set to constant (average volcano) background")
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.decoding
   :platform: Unix, Windows
   :synopsis: Represents an ontological decoding definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

class Decoding(object):
    """Represents information used for decoding from a non-standard representation, e.g. Metafor CIM XML.

    """

    def __init__(self, property_name, decoding, type):
        """Instance constructor.

        :param str property_name: Name of property with which decoding is associated.
        :param str decoding: Decoding information (e.g. an xpath expression).
        :param str|None type: Target type to be decoded (either a class or enum).

        """
        self.property_name = property_name
        self.decoding = decoding
        self.type = type



# -*- coding: utf-8 -*-

"""
.. module:: decoder_xml_utils.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: XML decoding utility functions.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework. @ 2013-08-28 14:41:13.340289.

"""
import uuid
import types

import arrow
from lxml import etree as et

from pyesdoc import exceptions
from pyesdoc.utils import convert



# Null uuid for checking whether one has to be generated.
NULL_UUID = u'00000000-0000-0000-0000-000000000000'



def _get_value(xml, rtype=unicode):
    """Returns xml node value."""
    # Get xml value.
    if isinstance(xml, types.ListType):
        xml = None if len(xml) == 0 else xml[0]
    if xml is None:
        return None

    # Get unicode.
    if rtype is unicode:
        if isinstance(xml, types.StringTypes):
            result = convert.str_to_unicode(xml)
        else:
            result = convert.str_to_unicode(et.tostring(xml))
    else:
        if isinstance(xml, types.StringTypes):
            result = convert.unicode_to_str(xml)
        else:
            result = et.tostring(xml)

    # Format.
    result = result.strip()
    result = result.rstrip('|')

    return result


def _convert_to_unicode(xml, nsmap=None):
    """Converts an etree element xml representation into a unicode type."""
    return _get_value(xml, str)


def _convert_to_bool(xml, nsmap=None):
    """Converts an etree element xml representation into a boolean type."""
    value = _get_value(xml)

    return True if value is not None and value.lower() in [u'true'] else bool()


def _convert_to_integer(xml, nsmap=None):
    """Converts an etree element xml representation into an integer type."""
    value = _get_value(xml)

    return int(value) if value is not None and value.upper() != u'NONE' else int()


def _convert_to_float(xml, nsmap=None):
    """Converts an etree element xml representation into a float type."""
    value = _get_value(xml)

    return float(value) if value is not None else float()


def _convert_to_uid(xml, nsmap=None):
    """Converts an etree element xml representation into a uid type."""
    value = _get_value(xml)
    if value is None or value == NULL_UUID:
        return uuid.uuid4()
    try:
        return uuid.UUID(value)
    except ValueError:
        return uuid.UUID(value[0:36])


def _convert_to_datetime(xml, nsmap=None):
    """Converts an etree element xml representation into a datetime type."""
    value = _get_value(xml)
    if value is None:
        return None
    if len(value) == 4:
        return arrow.get(value, u"YYYY").datetime
    else:
        return arrow.get(value).datetime


# Set of simple type convertors.
_SIMPLE_TYPE_DECODERS = {
    'bool' : _convert_to_bool,
    'date' : _convert_to_datetime,
    'datetime' : _convert_to_datetime,
    'datetime.date' : _convert_to_datetime,
    'datetime.datetime' : _convert_to_datetime,
    'float' : _convert_to_float,
    'int' : _convert_to_integer,
    'str' : _convert_to_unicode,
    'unicode' : _convert_to_unicode,
    'uri' : _convert_to_unicode,
    'uuid' : _convert_to_uid,
    'uuid.UUID' : _convert_to_uid,
}


def set_attributes(target, xml, nsmap, decodings):
    """Decodes entity attributes from a collection of decodings.

    :param object target: A pyesdoc object with a set of attributes to be assigned.
    :param lxml.etree._Element xml: An xml element.
    :param dict nsmap: Set of xml namespace mappings.
    :param dict decodings: Set of mappings used to perform decoding.

    :returns: A pyesdoc object with assigned attributes.
    :rtype: object

    """
    attrs = []

    # Iterate & apply decodings.
    for attr, is_iterable, type, xpath in \
        [d for d in decodings if len(d) == 4]:

        # Determine if this is a duplicate assignment.
        is_duplicate = attr in attrs
        if not is_duplicate:
            attrs.append(attr)

        # Determine if type is a simple one.
        is_simple_type = type in _SIMPLE_TYPE_DECODERS

        try:
            _set_attribute(target,
                           xml,
                           nsmap,
                           attr,
                           type,
                           xpath,
                           is_simple_type,
                           is_iterable,
                           is_duplicate)
        except Exception as e:
            msg = "\nES-DOC :: WARNING :: XML DECODING ERROR\n"
            msg += "\tTarget = {0};\n".format(target)
            msg += "\tAttribute name = {0};\n".format(attr)
            msg += "\tAttribute type = {0};\n".format(type)
            msg += "\tAttribute is iterable ? = {0};\n".format(is_iterable)
            msg += "\tAttribute is simple ? = {0};\n".format(is_simple_type)
            msg += "\tAttribute xpath = {0};\n".format(xpath)
            msg += "\tError = {0};\n".format(e)
            print msg

    # Support operation chaining.
    return target


def _set_attribute(target,
                   xml,
                   nsmap,
                   attr,
                   decoder,
                   xpath,
                   is_simple_type,
                   is_iterable,
                   is_duplicate):
    """Decodes entity attribute from a decoding."""
    # Escape if xpath is unassigned.
    if not xpath:
        return

    # Set target object / attribute.
    obj = target
    parts = attr.split('.')
    for i in range(len(parts) - 1):
        obj = getattr(obj, parts[i])
    att_name = parts[len(parts) - 1]

    # Get current attribute value.
    cur_value = getattr(obj, att_name)

    # Escape if processing a duplicate and the target is already assigned.
    if is_duplicate and not is_iterable and cur_value is not None:
        return

    # Format xpath when appropriate.
    if is_simple_type == True and \
       '@' not in xpath and \
       xpath.endswith('/text()') == False:
       xpath += '/text()'

    # Get attribute value.
    att_value = _get_attribute_value(xml, nsmap, decoder, xpath, is_simple_type, is_iterable)
    if is_iterable and isinstance(att_value, list):
        att_value = [v for v in att_value if v]

    # Set attribute value.
    setattr(obj, att_name,
            att_value if not is_iterable else cur_value + att_value)


def _get_attribute_value(xml, nsmap, decoder, xpath, is_simple_type, is_iterable):
    """Gets the value of an attribute from xml."""
    # Apply xpath (derive xml fragment from value is derived).
    att_xml = xml.xpath(xpath, namespaces=nsmap)
    if not att_xml:
        return [] if is_iterable else None

    # From xml derive value.
    # ... simple types.
    if is_simple_type:
        if decoder in _SIMPLE_TYPE_DECODERS:
            decoder = _SIMPLE_TYPE_DECODERS[decoder]
        if is_iterable:
            return [decoder(i, nsmap) for i in att_xml]
        else:
            decoded = decoder(att_xml, nsmap)
            return None if decoded == str() else decoded

    # ... complex types.
    else:
        return decode_xml(decoder, att_xml, nsmap, is_iterable)


def decode_xml(decoder, xml, nsmap, is_iterable):
    """Decodes either an entity or an entity collection from xml.

    :param function decoder: Decoder function pointer.
    :param lxml.etree._Element xml: An xml element.
    :param dict nsmap: Set of xml namespace mappings.
    :param bool take_first: Flag indicating whether to return only the first entity of a collection.

    :returns: Decoded entity or entity collection.
    :rtype: miscellaneous

    """
    # None if passed none.
    if xml is None:
        return None

    # Instance if passed etree element.
    if isinstance(xml, et._Element):
        return decoder(xml, nsmap)

    # Instance if passed etree element collection and caller wants first only.
    if isinstance(xml, types.ListType) and is_iterable == False:
        return None if len(xml) == 0 else decode_xml(decoder, xml[0], nsmap, None)

    # Collection if passed etree element collection.
    if isinstance(xml, types.ListType):
        collection = []
        for elem in xml:
            instance = decode_xml(decoder, elem, nsmap, None)
            collection.append(instance)
        return collection

    # otherwise exception
    raise exceptions.DecodingException("xml cannot be decoded.")


def load_xml(xml, return_nsmap=False, default_ns='cim'):
    """Loads etree xml element.

    :param string xml: An xml blob.
    :param bool return_nsmap: Flag indicating whether namespace map will be returned or not.
    :param str default_ns: Default namespace.

    :returns: XML element.
    :rtype: lxml.etree._Element

    """
    # Defensive programming.
    if xml is None:
        raise exceptions.DecodingException("XML is undefined.")

    # ... etree elements.
    nsmap = None
    if isinstance(xml, et._Element):
        nsmap = xml.nsmap

    # ... etree element trees.
    elif isinstance(xml, et._ElementTree):
        xml = xml.getroot()
        nsmap = xml.nsmap

    else:
        # ... files / url's
        try:
            xml = et.parse(xml)
            xml = xml.getroot()
            nsmap = xml.nsmap
        except Exception as e:
            # ... unicode
            if isinstance(xml, unicode):
                xml = convert.unicode_to_str(xml)

            # ... strings
            if isinstance(xml, str):
                try:
                    xml = et.fromstring(xml)
                except Exception as err:
                    raise exceptions.DecodingException("Invalid xml string.")
                else:
                    nsmap = xml.nsmap

            # Unsupported
            else:
                raise exceptions.DecodingException("Unsupported xml type, must be either a string, file, url or etree.")


    # Set default namespace.
    if nsmap is not None:
        nsmap[default_ns] = nsmap.pop(None)

    # Return either a tuple or single.
    if return_nsmap:
        return xml, nsmap
    else:
        return xml



# -*- coding: utf-8 -*-

"""
.. module:: platform_classes.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""

def component_performance():
    """Describes the simulation rate of a model component.

Based on "CPMIP: Measurements of Real Computational Performance of
Earth System Models" (Balaji et. al. 2016, doi:10.5194/gmd-2016-197,
http://www.geosci-model-dev-discuss.net/gmd-2016-197/)

    """
    return {
        'type': 'class',
        'base': 'platform.performance',
        'is_abstract': False,
        'pstr': ('{} (sypd:{})', ('name', 'sypd')),
        'properties': [            
            ('component', 'software.software_component', '1.1',
             "Link to a CIM software component description."),
        ],
        'constraints': [
            ('cardinality', 'model', '0.0'),
            ('cardinality', 'resolution', '0.0'),
            ('cardinality', 'complexity', '0.0'),
            ('cardinality', 'platform', '0.0'),
            ('cardinality', 'compiler', '0.0'),
            ('cardinality', 'joules_per_simulated_year', '0.0'),
            ('cardinality', 'coupling_cost', '0.0'),
            ('cardinality', 'memory_bloat', '0.0'),
            ('cardinality', 'data_output_cost', '0.0'),
            ('cardinality', 'data_instensity', '0.0'),
        ],
    }

def compute_pool():
    """Homogeneous pool of nodes within a computing machine.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('accelerator_type', 'str', '0.1',
                "Type of accelerator."),
            ('accelerators_per_node', 'int', '0.1',
                "Number of accelerator units on a node."),
            ('clock_cycle_concurrency', 'int', '0.1',
                 'The number of operations which can be carried out concurrently in a single clock cycle of a single core. E.g. 4.'),
            ('clock_speed', 'float', '0.1',
                 'The clock speed of a single core, in units of GHz. E.g. 3.6.'),
            ('compute_cores_per_node', 'int', '0.1',
                "Number of CPU cores per node."),
            ('cpu_type', 'str', '0.1',
                "CPU type."),
            ('description', 'str', '0.1',
                "Textural description of pool."),
            ('interconnect', 'str', '0.1',
                "Interconnect used."),
            ('memory_per_node', 'platform.storage_volume', '0.1',
                "Memory per node."),
            ('model_number', 'str', '0.1',
                "Model/Board number/type."),
            ('name', 'str', '0.1',
                "Name of compute pool within a machine."),
            ('number_of_nodes', 'int', '0.1',
                "Number of nodes."),
            ('operating_system', 'str', '0.1',
                "Operating system.")
        ],
        'derived': [
            ('total_cores', 'compute_cores_per_node * number_of_nodes'),
            ('total_memory', 'memory_per_node * number_of_nodes')
        ]
    }


def machine():
    """A computer/system/platform/machine which is used for simulation.

    """
    return {
        'type': 'class',
        'base': 'platform.partition',
        'is_abstract': False,
        'is_document': True,
        'properties': [

        ]
    }


def partition():
    """A major partition (component) of a computing system (aka machine).

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('compute_pools', 'platform.compute_pool', '1.N',
                "Layout of compute nodes."),
            ('description', 'str', '0.1',
                "Textural description of machine."),
            ('institution', 'linked_to(shared.party)', '1.1',
                "Institutional location."),
            ('model_number', 'str', '0.1',
                "Vendor's model number/name - if it exists."),
            ('name', 'str', '1.1',
                "Name of partition (or machine)."),
            ('online_documentation', 'shared.online_resource', '0.N',
                "Links to documentation."),
            ('partition', 'platform.partition', '0.N',
                "If machine is partitioned, treat subpartitions as machines."),
            ('storage_pools', 'platform.storage_pool', '0.N',
                "Storage resource available."),
            ('vendor', 'linked_to(shared.party)', '0.1',
                "The system integrator or vendor."),
            ('when_used', 'time.time_period', '0.1',
                "If no longer in use, the time period it was in use.")
        ]
    }

def performance():
    """Describes the properties of a performance of a configured model on
a particular system/machine.

Based on "CPMIP: Measurements of Real Computational Performance of
Earth System Models" (Balaji et. al. 2016, doi:10.5194/gmd-2016-197,
http://www.geosci-model-dev-discuss.net/gmd-2016-197/)

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{} (sypd:{})', ('name', 'sypd')),
        'is_document': True,
        'properties': [
            ('name', 'str', '0.1',
                "Name for performance (experiment/test/whatever)."),
            
            # CPMIP model and platform
            ('model', 'science.model', '1.1',
                "Model for which performance was tested."),
            ('resolution', 'int', '0.1', 
                 'Resolution measured as the number of gridpoints (or more generally, spatial degrees of freedom) NX x NY x NZ per component with an independent discretization'),
            ('complexity', 'int', '0.1', 
                 'Complexity measured as the number of prognostic variables per component with an independent discretization'),
            ('platform', 'platform.machine', '1.1',
                 'Platform on which performance was tested.'),
            ('compiler', 'str', '0.1',
                "Compiler used for performance test."),

            # CPMIP computational cost
            ('simulated_years_per_day', 'float', '0.1',
                 'Simulated years per day (SYPD) in a 24h period on the given platform'),
            ('actual_simulated_years_per_day', 'float', '0.1',
                 'Actual simulated years per day (ASYPD) in a 24h period on the given platform obtained from a typical long-running simulation with the model. Inclusive of system interruptions, queue wait time, or issues with the model workflow, etc.'),
            ('core_hours_per_simulated_year', 'float', '0.1',
                 'Core-hours per simulated year (CHSY). This is measured as the product of the model runtime for 1 SY, and the numbers of cores allocated. Note that if allocations are done on a node basis then all cores on a node are charged against the allocation, regardless of whether or not they are used.'),
            ('parallelization', 'float', '0.1',
                 'Parallelization measured as the total number of cores (NP) allocated for the run, regardless of whether or or all cores were used. Note that NP=CHSY*SYPD/24.'),
            ('joules_per_simulated_year', 'float', '0.1',
                 'The energy cost of a simulation, measured in joules per simulated year (JPSY). Given the energy E in joules consumed over a budgeting interval T (generally 1 month or 1 year, in units of hours), JPSY=CHSY*E*T/NP'),

            # CPMIP coupling, memory, I/O
            ('coupling_cost', 'float', '0.1',
                 'Coupling cost measures the overhead caused by coupling. This can include the cost of the coupling algorithm itself (which may involve grid interpolation and computation of transfer coefficients for conservative coupling) as well as load imbalance. It is the normalized difference between the time-processor integral for the whole model versus the sum of individual concurrent components'),
            ('memory_bloat', 'float', '0.1',
                 'Memory bloat is the ratio of the actual memory size to the ideal memory size (the size of the complete model state, which in theory is all you need to hold in memory)Mi, defined below.'),
            ('data_output_cost', 'float', '0.1',
                 'Data output cost is the cost of performing I/O, and is the ratio of CHSY with and without I/O.'),
            ('data_intensity', 'float', '0.1',
                 'Data intensity the amount of data produced per compute-hour, in units GB per compute-hour.'),

            # Subcomponent performance
            ('subcomponent_performance', 'platform.component_performance', '0.N',
                "Describes the performance of each subcomponent.")
        ]
    }

def storage_pool():
    """Homogeneous storage pool on a computing machine.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('description', 'str', '0.1',
                "Description of the technology used."),
            ('name', 'str', '1.1',
                "Name of storage pool."),
            ('type', 'platform.storage_systems', '0.1',
                "Type of storage."),
            ('vendor', 'linked_to(shared.party)', '0.1',
                "Vendor of storage hardware.")
        ]
    }


def storage_systems():
    """Controlled vocabulary for storage  types (including filesystems).

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("Lustre", "Lustre parallel file system"),
            ("GPFS", "IBM GPFS (also known as IBM Spectral Scale"),
            ("isilon", "The EMC scaleout NAS solution"),
            ("NFS", "Generic Network File System"),
            ("S3", "Object file system exposing the AWS S3 interface"),
            ("PanFS", "Panasas Parallel File system"),
            ("Other Disk", "Other disk based file system"),
            ("Tape - MARS", "Tape storage system using ECMWF MARS"),
            ("Tape - MASS", "Tape storage system using Met Office MASS"),
            ("Tape - Castor", "Tape storage sytsem using CERN Castor"),
            ("Tape - Other", "Other tape based system")
        ]
    }


def storage_volume():
    """Platform storage volume and units.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{} {}', ('volume', 'units')),
        'properties': [
            ('units', 'platform.volume_units', '1.1',
                "Volume units."),
            ('volume', 'int', '1.1',
                "Numeric value.")
        ]
    }


def volume_units():
    """Appropriate storage volume units.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("GB", "Gigabytes (1000^3)"),
            ("TB", "Terabytes (1000^4)"),
            ("PB", "Petabytes (1000^5)"),
            ("EB", "Exabytes (1000^6)"),
            ("TiB", "Tebibytes (1024^4)"),
            ("PiB", "Pebibytes (1024^5)"),
            ("EiB", "Exbibytes (1024^6)")
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.enum
   :platform: Unix, Windows
   :synopsis: Represents an ontological enumerated type definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

class Enum(object):
    """Represents an enumeration, i.e. a constrained set of values.

    """
    def __init__(self, name, is_open, doc_string, members):
        """Instance constructor.

        :param str name: Enumeration name.
        :param bool is_open: Flag indicating whether members can be added to the enumeration or not.
        :param str doc_string: Enumeration documentation string.
        :param iterable members: Set of associated enumeration members.

        """
        self.name = name
        self.is_open = is_open
        self.doc_string = doc_string if doc_string is not None else ''
        self.members = members
        self.package = None

        self.op_doc_string_name = None
        self.op_file_name = None
        self.op_full_name = None
        self.op_func_name = None
        self.op_import_name = None
        self.op_name = None


    def __repr__(self):
        """Instance string representation.

        """
        try:
            return "{}.{}".format(self.package, self.name)
        except:
            return self.name




# -*- coding: utf-8 -*-

from pyesdoc._extensions.cim.v1 import activity_ensemble
from pyesdoc._extensions.cim.v1 import activity_numerical_experiment
from pyesdoc._extensions.cim.v1 import activity_simulation_run
from pyesdoc._extensions.cim.v1 import data_data_object
from pyesdoc._extensions.cim.v1 import grids_grid_spec
from pyesdoc._extensions.cim.v1 import misc_document_set
from pyesdoc._extensions.cim.v1 import software_model_component
from pyesdoc._extensions.cim.v1 import quality_cimquality



# Supported extenders keyed by document type.
SUPPORTED = {
	"cim.1.activity.ensemble": activity_ensemble,
	"cim.1.activity.numericalexperiment": activity_numerical_experiment,
	"cim.1.activity.simulationrun": activity_simulation_run,
	"cim.1.data.dataobject": data_data_object,
	"cim.1.grids.gridspec": grids_grid_spec,
	"cim.1.misc.documentset": misc_document_set,
	"cim.1.software.modelcomponent": software_model_component,
	"cim.1.software.statisticalmodelcomponent": software_model_component,
	"cim.1.quality.cimquality": quality_cimquality,
}



# -*- coding: utf-8 -*-

"""
.. module:: designing_project.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.designing.project document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections

from pyesdoc import constants
from pyesdoc.ontologies import cim



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_viewer_urls,
        _set_summary_fields,
        _set_tiered_experiments,
        _set_other_experiments
        )


def _set_viewer_urls(ctx):
    """Sets viewer urls.

    """
    for exp in ctx.doc.required_experiments + ctx.doc.governed_experiments:
        if isinstance(exp, cim.v2.designing.NumericalExperiment):
            exp.viewer_url = constants.VIEWER_URL_BY_ID.format(ctx.meta.project, exp.meta.id, exp.meta.version)
        elif isinstance(exp, cim.v2.shared.DocReference):
            exp.viewer_url = constants.VIEWER_URL_BY_ID.format(ctx.meta.project, exp.id, exp.version)

    for p in ctx.doc.sub_projects:
        if isinstance(p, cim.v2.designing.Project):
            p.viewer_url = constants.VIEWER_URL_BY_ID.format(ctx.meta.project, p.meta.id, p.meta.version)
        elif isinstance(p, cim.v2.shared.DocReference):
            p.viewer_url = constants.VIEWER_URL_BY_ID.format(ctx.meta.project, p.id, p.version)


def _set_summary_fields(ctx):
    """Sets related experiment viewer urls.

    """
    ctx.ext.summary_fields += (
        ctx.doc.canonical_name,
        ctx.doc.name,
        ctx.doc.long_name
    )


def _set_tiered_experiments(ctx):
    """Groups governed experiments by tier.

    """
    tiered_experiments = collections.defaultdict(list)
    for e in ctx.doc.governed_experiments:
        if e.further_info is not None:
            tiered_experiments[e.further_info].append(e)
    ctx.ext.tiered_experiments = tiered_experiments


def _set_other_experiments(ctx):
    """Sets non-governed experiments.

    """
    governed_experiments = [j.name for j in ctx.doc.governed_experiments]
    ctx.ext.other_experiments = [i for i in ctx.doc.required_experiments
                                 if i.name not in governed_experiments]




# -*- coding: utf-8 -*-

"""
.. module:: software_enums.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""


def coupling_framework():
    """The set of terms which define known coupling frameworks.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("OASIS", "The OASIS coupler - prior to OASIS-MCT"),
            ("OASIS3-MCT", "The MCT variant of the OASIS coupler"),
            ("ESMF", "Vanilla Earth System Modelling Framework"),
            ("NUOPC", "National Unified Operational Prediction Capability variant of ESMF"),
            ("Bespoke", "Customised coupler developed for this model"),
            ("Unknown", "It is not known what/if-a coupler is used"),
            ("None", "No coupler is used")
        ]
    }


def programming_language():
    """The set of terms which define programming languages used for earth
    system simulation.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("Fortran", "Fortran Programming Language"),
            ("C", "C Programmming Language"),
            ("C++", "C++ Programming Language"),
            ("Python", "Python Programming Language")
        ]
    }



"""

A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Ruth Petrie'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Ruth Petrie, Bryan Lawrence'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Sea Ice Thermodynamics'

# --------------------------------------------------------------------
# SUB-PROCESS: Processes related to energy in sea ice thermodynamics
# --------------------------------------------------------------------
DETAILS['energy'] = {
    'description': 'Processes related to energy in sea ice thermodynamics',
    'properties': [
        ('enthalpy_formulation', 'ENUM:energy_formulation', '1.1',
            "What is the energy formulation?"),
        ('thermal_conductivity', 'ENUM:thermal_conductivity_type', '1.1',
            "What type of thermal conductivity is used?"),
        ('heat_diffusion', 'ENUM:heat_diffusion_type', '1.1',
            "What is the method of heat diffusion?"),
        ('basal_heat_flux', 'ENUM:basal_heat_flux_method', '1.1',
            "Method by which basal ocean heat flux is handled?"),
        ('fixed_salinity_value', 'float', '0.1',
             "If you have selected {Thermal properties depend on S-T (with fixed salinity)}, supply fixed salinity value for each sea ice layer."),
        ('heat_content_precip', 'str', '1.1',
            "Describe the method by which the heat content of precipitation is handled.")
        ]
    }


# --------------------------------------------------------------------
# SUB-PROCESS: Mass.
# --------------------------------------------------------------------
DETAILS['mass'] = {
    'description': 'Processes related to mass in sea ice thermodynamics',
    'properties': [
        ('new_ice_formation', 'str', '1.1',
             "Describe the method by which new sea ice is formed in open water."),
        ('ice_vertical_growth_and_melt', 'str', '1.1',
             "Describe the method that governs the vertical growth and melt of sea ice."),
        ('ice_lateral_melting', 'ENUM:lateral_melting_types', '1.1',
             "What is the method of sea ice lateral melting?"),
        ('ice_surface_sublimation', 'str', '1.1',
             "Describe the method that governs sea ice surface sublimation."),
        ('frazil_ice', 'str', '1.1',
             "Describe the method of frazil ice formation."),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Salt.
# --------------------------------------------------------------------
DETAILS['salt'] = {
    'description': 'Processes related to salt in sea ice thermodynamics.',
    'properties': [
        ('has_multiple_sea_ice_salinities', 'bool', '1.1',
            "Does the sea ice model use two different salinities: one for thermodynamic calculations; and one for the salt budget?"),
        ('sea_ice_salinity_thermal_impacts', 'bool', '1.1',
            "Does sea ice salinity impact the thermal properties of sea ice?"),
    ],
    'detail_sets': [
        'mass_transport',
        'thermodynamics',
    ]
}

DETAILS['salt:mass_transport']={
    'description':'Mass transport of salt',
    'properties': [
        ('salinity_type', 'ENUM:salinity_method', '1.1',
            "How is salinity determined in the mass transport of salt calculation?"),
        ('constant_salinity_value', 'float', '0.1',
             "If using a constant salinity value specify this value in PSU?"),
        ('additional_details', 'str', '0.1',
            "Describe the salinity profile used."),
        ]
    }

DETAILS['salt:thermodynamics']={
    'description':'Salt thermodynamics',
    'properties': [
        ('salinity_type', 'ENUM:salinity_method', '1.1',
            "How is salinity determined in the thermodynamic calculation?"),
        ('constant_salinity_value', 'float', '0.1',
             "If using a constant salinity value specify this value in PSU?"),
        ('additional_details', 'str', '0.1',
            "Describe the salinity profile used."),
        ]
    }


# --------------------------------------------------------------------
# SUB-PROCESS: Ice Thickness Distribution.
# --------------------------------------------------------------------
DETAILS['ice_thickness_distribution'] = {
    'description': 'Ice thickness distribution details.',
    'properties': [
        ('representation', 'ENUM:ice_thickness_representation', '1.1',
            "How is the sea ice thickness distribution represented?"),
        ]
    }


# --------------------------------------------------------------------
# SUB-PROCESS: Melt Ponds
# --------------------------------------------------------------------
DETAILS['melt_ponds'] = {
    'description': 'Characteristics of melt ponds.',
    'properties': [
        ('are_included', 'bool', '1.1',
            "Are melt ponds included in the sea ice model?"),
        ('formulation', 'ENUM:melt_pond_formulation', '1.1',
            "What method of melt pond formulation is used?"),
        ('impacts', 'ENUM:melt_pond_impacts', '1.N',
            "What do melt ponds have an impact on?")
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Snow thermodynamic processes.
# --------------------------------------------------------------------
DETAILS['snow_processes'] = {
    'description': 'Thermodynamic processes in snow on sea ice',
    'properties': [
        ('has_snow_aging', 'bool', '1.N', "Set to True if the sea ice model has a snow aging scheme."),
        ('snow_aging_scheme', 'str', '0.1', "Describe the snow aging scheme."),
        ('has_snow_ice_formation', 'bool', '1.N', "Set to True if the sea ice model has snow ice formation."),
        ('snow_ice_formation_scheme', 'str', '0.1', "Describe the snow ice formation scheme."),
        ('redistribution', 'str', '1.1', "What is the impact of ridging on snow cover?"),
        ('heat_diffusion', 'ENUM:snow_process_types', '1.1',
            "What is the heat diffusion through snow methodology in sea ice thermodynamics?"),
    ]
}

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['energy_formulation'] = {
    'description': 'Thermodynamic energy formulations',
    'is_open': True,
    'members': [
        ('Pure ice latent heat (Semtner 0-layer)', None),
        ('Pure ice latent and sensible heat', None),
        ('Pure ice latent and sensible heat + brine heat reservoir (Semtner 3-layer)', None),
        ('Pure ice latent and sensible heat + explicit brine inclusions (Bitz and Lipscomb)', None),
    ]
}

ENUMERATIONS['thermal_conductivity_type'] = {
    'description': 'Types of thermal conductivity',
    'is_open': True,
    'members': [
        ('Pure ice', None),
        ('Saline ice', None),
    ]
}

ENUMERATIONS['basal_heat_flux_method'] = {
    'description': 'Basal ocean heat flux methodology',
    'is_open': True,
    'members': [
        ('Heat Reservoir', 'Brine inclusions treated as a heat reservoir'),
        ('Thermal Fixed Salinity', 'Thermal properties depend on S-T (with fixed salinity)'),
        ('Thermal Varying Salinity', 'Thermal properties depend on S-T (with varying salinity'),
    ]
}

ENUMERATIONS['heat_diffusion_type'] = {
    'description': 'Heat diffusion types',
    'is_open': True,
    'members': [
        ('Conduction fluxes', None),
        ('Conduction and radiation heat fluxes', None),
        ('Conduction, radiation and latent heat transport', None),
    ]
}

ENUMERATIONS['lateral_melting_types'] = {
    'description': 'Ice lateral melting methods',
    'is_open': True,
    'members': [
        ('Floe-size dependent (Bitz et al 2001)', None),
        ('Virtual thin ice melting (for single-category)', None),
    ]
}

ENUMERATIONS['ice_thickness_representation'] = {
    'description': 'Types of sea ice thickness representation',
    'is_open': True,
    'members': [
        ('Explicit', None),
        ('Virtual (enhancement of thermal conductivity, thin ice melting)', None),
    ]
}

ENUMERATIONS['thermo_brine_types'] = {
    'description': 'Brine inclusion methodologies',
    'is_open': True,
    'members': [
        ('Heat Reservoir', 'Brine inclusions treated as a heat reservoir'),
        ('Thermal Fixed Salinity', 'Thermal properties depend on S-T (with fixed salinity)'),
        ('Thermal Varying Salinity', 'Thermal properties depend on S-T (with varying salinity'),
    ]
}

ENUMERATIONS['melt_pond_formulation'] = {
    'description': 'Formulations of melt ponds',
    'is_open': True,
    'members': [
        ('Flocco and Feltham (2010)', None),
        ('Level-ice melt ponds', None),
    ]
}

ENUMERATIONS['melt_pond_impacts'] = {
    'description': 'Impacts of melt ponds',
    'is_open': True,
    'members': [
        ('Albedo', None),
        ('Freshwater', None),
        ('Heat', None),
    ]
}

ENUMERATIONS['snow_process_types'] = {
    'description': 'Types of snow processes',
    'is_open': True,
    'members': [
        ('single-layered heat diffusion', None),
        ('multi-layered heat diffusion', None),
    ]
}

ENUMERATIONS['salinity_method'] = {
    'description': 'Method of describing the value of salinity',
    'is_open': True,
    'members': [
        ('constant', None),
        ('prescribed salinity profile', None),
        ('prognostic salinity profile', None),
    ]
}




# -*- coding: utf-8 -*-

"""
.. module:: file_info.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
    :synopsis: An archived document folder wrapper.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import glob
import os

import pyesdoc
from pyesdoc.archive import constants
from pyesdoc.archive.file_info import ArchiveFileInfo
from pyesdoc.constants import DEFAULT_ENCODING
from pyesdoc._extensions import extend as extend_doc
from pyesdoc._io import read as read_doc



def _get_file_filterset():
    """Returns set of supported file filters.

    """
    result = constants.FILE_FILTER_SET
    for doc_type in pyesdoc.ontologies.get_doc_type_keyset():
        result.add("{}_*.*".format(doc_type.replace(".", "-").lower()))

    return result


class ArchiveFolderInfo(object):
    """An archive folder.

    """
    def __init__(self, project, source, path):
        """Instance constructor.

        """
        self.path = path
        self.project = project.lower()
        self.source = source.lower()
        self._paths = \
            { ff: os.path.join(path, ff) for ff in _get_file_filterset() }


    def __repr__(self):
        """Instance representation.

        """
        return "{0}".format(self.path)


    @property
    def exists(self):
        """Gets flag indicating whether folder exists.

        """
        return os.path.exists(self.path)


    def delete(self):
        """Deletes underlying folder from local file system.

        """
        if self.exists:
            os.remove(self.path)


    def get_count(self, file_filter=None):
        """Returns file count.

        """
        if not file_filter:
            file_filter = constants.FILE_FILTER_ALL
        return len(glob.glob(self._paths[file_filter]))


    def yield_documents(self, file_filter=None):
        """Yields set of documents.

        :param str file_filter: Optional file filter.

        :returns: Set of file paths.
        :rtype: generator

        """
        for file_info in self.yield_files(file_filter):
            yield extend_doc(read_doc(file_info.path))


    def yield_files(self, file_filter=None):
        """Yields set of files.

        :param str file_filter: Optional file filter.

        :returns: Set of file paths.
        :rtype: generator

        """
        if file_filter is None or file_filter == "*":
            file_filter = constants.FILE_FILTER_ALL
        elif file_filter != constants.FILE_FILTER_ERROR and file_filter.find("_") == -1:
            file_filter = "{}_*.*".format(file_filter)
        for fpath in glob.iglob(self._paths[file_filter]):
            yield ArchiveFileInfo(self, fpath)


    def find_file(self, uid, version):
        """Yields files for processing.

        :param str uid: Document unique identifier.
        :param str version: Document version.

        :returns: A file wrapper to be processed.
        :rtype: pyesdoc.archive.ArchiveFileInfo

        """
        try:
            int(version)
        except ValueError:
            file_filter = "*_{}_*.*".format(uid)
        else:
            file_filter = "*_{}_{}_*.*".format(uid, version)
        files = glob.glob(os.path.join(self.path, file_filter))
        if files:
            return ArchiveFileInfo(self, sorted(files)[-1])


    def write(self, doc):
        """Writes a document to the archive.

        :param object doc: Document to be written.

        :returns: Pointer to archive file.
        :rtype: pyesdoc.archive.ArchiveFileInfo

        """
        # Ensure document is extended.
        extend_doc(doc)

        # Set file path.
        fname = "{}_{}_{}.{}".format(doc.meta.type, doc.meta.id, doc.meta.version, DEFAULT_ENCODING)
        fpath = os.path.join(self.path, fname)

        # Write to file system.
        with open(fpath, 'w') as fstream:
            fstream.write(pyesdoc.encode(doc, DEFAULT_ENCODING))

        # Return file wrapper.
        return ArchiveFileInfo(self, fpath)



"""A realm grid sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmosphere grid'

# --------------------------------------------------------------------
# DISCRETISATION: Description of the horizontal discretization and numerics.
# --------------------------------------------------------------------
DETAILS['discretisation'] = {
    'description': 'Atmosphere grid discretisation',
    'properties': [],
    'detail_sets': [
        'horizontal',
        'vertical',
        ]
    }

DETAILS['discretisation:horizontal'] = {
    'description': 'Atmosphere discretisation in the horizontal',
    'properties': [
        ('scheme_type', 'ENUM:dynamical_core_discretisation_horizontal_type', '1.1',
            'Horizontal discretisation type'),
        ('scheme_method', 'ENUM:dynamical_core_discretisation_horizontal_method', '1.1',
            'Horizontal discretisation method'),
        ('scheme_order', 'int', '1.1',
            'Horizontal discretisation function order'),
        ('horizontal_pole', 'ENUM:dynamical_core_discretisation_horizontal_pole', '1.1',
            'Horizontal discretisation pole singularity treatment'),
        ]
    }

DETAILS['discretisation:vertical'] = {
    'description': 'Atmosphere discretisation in the vertical',
    'properties': [
        ('coordinate_type', 'ENUM:dynamical_core_discretisation_vertical_type', '1.N',
            'Type of vertical coordinate system'),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['dynamical_core_discretisation_horizontal_type'] = {
    'description': 'Type of horizontal discretisation scheme',
    'is_open': True,
    'members': [
        ('spectral', None),
        ('fixed grid', None),
        ]
    }

ENUMERATIONS['dynamical_core_discretisation_horizontal_method'] = {
    'description': 'If the scheme type is fixed grid, describe the scheme method for the horizontal discretisation scheme',
    'is_open': False,
    'members': [
        ('finite elements', None),
        ('finite volumes', None),
        ('finite difference', None),
        ('centered finite difference', None),
        ]
    }

ENUMERATIONS['dynamical_core_discretisation_horizontal_pole'] = {
    'description': 'Method used to handle the pole singularities',
    'is_open': True,
    'members': [
        ('filter', None),
        ('pole rotation', None),
        ('artificial island', None),
        ]
    }

ENUMERATIONS['dynamical_core_discretisation_vertical_type'] = {
    'description': 'Type of vertical coordinate system',
    'is_open': True,
    'members': [
        ('isobaric', 'vertical coordinate on pressure levels'),
        ('sigma', 'allows vertical coordinate to follow model terrain'),
        ('hybrid sigma-pressure', 'sigma system near terrain and isobaric above'),
    ]
}


# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_activity_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.activity package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_shared_package as shared



class Activity(object):
    """An abstract class within the cim v2 type system.

    Base class for activities.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(Activity, self).__init__()

        self.alternative_names = []                       # unicode (0.N)
        self.canonical_name = None                        # unicode (0.1)
        self.citations = []                               # shared.Citation (0.N)
        self.description = None                           # unicode (0.1)
        self.duration = None                              # time.TimePeriod (0.1)
        self.internal_name = None                         # unicode (0.1)
        self.keywords = None                              # unicode (0.1)
        self.long_name = None                             # unicode (0.1)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.name = None                                  # unicode (1.1)
        self.previously_known_as = []                     # unicode (0.N)
        self.rationale = None                             # unicode (0.1)
        self.responsible_parties = []                     # shared.Responsibility (0.N)


class AxisMember(object):
    """A concrete class within the cim v2 type system.

    Description of a given ensemble member. It will normally be related to a specific
    ensemble requirement. Note that start dates can be extracted directly from the simulations
    and do not need to be recorded with an axis member description.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(AxisMember, self).__init__()

        self.description = None                           # unicode (1.1)
        self.extra_detail = None                          # unicode (0.1)
        self.index = None                                 # int (1.1)
        self.value = None                                 # float (0.1)


class EnsembleAxis(object):
    """A concrete class within the cim v2 type system.

    Defines the meaning of r/i/p indices in an ensemble.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(EnsembleAxis, self).__init__()

        self.extra_detail = None                          # unicode (1.1)
        self.member = []                                  # activity.AxisMember (1.N)
        self.short_identifier = None                      # unicode (1.1)
        self.target_requirement = None                    # designing.NumericalRequirement (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.axis)


class EnsembleMember(object):
    """A concrete class within the cim v2 type system.

    An ensemble may be a complicated interplay of axes, for example, r/i/p, not all of which
    are populated, so we need a list of the actual simulations and how they map onto the ensemble
    axes.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(EnsembleMember, self).__init__()

        self.errata = None                                # shared.OnlineResource (0.1)
        self.had_performance = None                       # platform.Performance (0.1)
        self.ran_on = None                                # platform.Machine (0.1)
        self.simulation = None                            # data.Simulation (1.1)


class ParentSimulation(object):
    """A concrete class within the cim v2 type system.

    Defines the relationship between a simulation and its parent.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(ParentSimulation, self).__init__()

        self.branch_method = None                         # unicode (0.1)
        self.branch_time_in_child = None                  # time.DateTime (0.1)
        self.branch_time_in_parent = None                 # time.DateTime (0.1)
        self.parent = None                                # data.Simulation (1.1)


class Conformance(Activity):
    """A concrete class within the cim v2 type system.

    A specific conformance. Describes how a particular numerical requirement has been implemented.
    Will normally be linked from an ensemble descriptor.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Conformance, self).__init__()

        self.conformance_achieved = None                  # activity.ConformanceType (1.1)
        self.datasets = []                                # data.InputDataset (0.N)
        self.models = []                                  # science.Model (1.N)
        self.target_requirement = None                    # designing.NumericalRequirement (1.1)


class Ensemble(Activity):
    """A concrete class within the cim v2 type system.

    Generic ensemble definition.
    Holds the definition of how the various ensemble members have been configured.
    If ensemble axes are not present, then this is either a single member ensemble,
    or part of an uber ensemble.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Ensemble, self).__init__()

        self.common_conformances = []                     # activity.Conformance (0.N)
        self.common_performance = None                    # platform.Performance (0.1)
        self.documentation = []                           # shared.OnlineResource (0.N)
        self.ensemble_axes = []                           # activity.EnsembleAxis (0.N)
        self.experiments = []                             # designing.NumericalExperiment (1.N)
        self.members = []                                 # activity.EnsembleMember (1.N)
        self.uber_ensembles = []                          # activity.UberEnsemble (0.N)


class UberEnsemble(Ensemble):
    """A concrete class within the cim v2 type system.

    An ensemble made up of other ensembles. Often used where parts of an ensemble were run by
    different institutes. Could also be used when a new experiment is designed which can use
    ensemble members from previous experiments and/or projects.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(UberEnsemble, self).__init__()

        self.child_ensembles = []                         # activity.Ensemble (1.N)


class ConformanceType(object):
    """An enumeration within the cim v2 type system.

    Standardised set of conformance responses.
    """
    is_open = True
    members = [
        "Conformed",
        "Partially Conformed",
        "Not Conformed",
        "Not Applicable"
        ]
    descriptions = [
        "Simulation (or ensemble) conformed to requirement",
        "Simulation (or ensemble) partially conformed to requirement - details in description",
        "Simulation (or ensemble) was unable to conform to requirement",
        "Requirement is not relevant"
        ]





# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.schema_validation.context
   :platform: Unix, Windows
   :synopsis: Contextual information passed to a validator.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import inspect



class ValidationContext(object):
    """Encapsulates schema validation processing information.

    """
    def __init__(self, schema):
        """Instance constructor.

        """
        self.schema = schema
        self.report = list()


    def set_error(self, err):
        """Adds an error to the manged collection.

        """
        self.report.append(err)


    @property
    def schema_name(self):
        """Gets schema name.

        """
        return self.schema.NAME


    @property
    def schema_version(self):
        """Gets schema version.

        """
        return "v{}".format(self.schema.VERSION)


    @property
    def package_factories(self):
        """Gets package factories.

        """
        return self.get_functions(self.schema)


    @property
    def packages(self):
        """Gets package definitions.

        """
        result = list()
        for factory in self.package_factories:
            try:
                result.append((factory, factory()))
            except Exception as err:
                pass

        return sorted([p for p in result if isinstance(p[1], set) and len(p[1])])


    @property
    def type_modules(self):
        """Gets type modules.

        """
        result = list()
        for factory, modules in self.packages:
            result += [(factory, m) for m in modules
                       if inspect.ismodule(m) and self.get_functions(m)]

        return sorted(result)


    @property
    def type_factories(self):
        """Gets type factories.

        """
        result = list()
        for factory, module in self.type_modules:
            result += [(module, f) for f in self.get_functions(module)]

        return sorted(result)


    @property
    def types(self):
        """Gets type definitions.

        """
        result = list()
        for module, factory in self.type_factories:
            try:
                type_ = factory()
            except Exception as err:
                pass
            else:
                if isinstance(type_, dict):
                    result.append((module, factory, type_))

        return sorted(result)


    @property
    def classes(self):
        """Get class definitions.

        """
        return [(t[0], t[1], t[2]) for t in self.types
                if 'type' in t[2] and t[2]['type'] == 'class']


    @property
    def enums(self):
        """Get enum definitions.

        """
        return [(t[0], t[1], t[2]) for t in self.types
                if 'type' in t[2] and t[2]['type'] == 'enum']


    def get_name(self, factory, module=None):
        """Gets a name used when displaying an error message.

        """
        if module is None or module == self.schema:
            return "{0}.v{1}.{2}".format(
                self.schema.NAME,
                self.schema.VERSION,
                factory.__name__)
        else:
            return "{0}.v{1}.{2}.{3}".format(
                self.schema.NAME,
                self.schema.VERSION,
                self.get_package_name(module),
                factory.__name__)


    def get_package_name(self, module):
        """Returns name of a package from a module.

        """
        return module.__name__.split('.')[-1].split('_')[0]


    def get_class_property_types(self, cls):
        """Returns class property type definitions.

        """
        return [(p[0], p[1]) for p in cls.get('properties', [])
                if len(p[1].split(".")) > 1]


    def get_type_name(self, factory, module):
        """Gets a type name.

        """
        return "{0}.{1}".format(
            self.get_package_name(module),
            factory.__name__)


    def get_functions(self, module):
        """Returns a collection of function pointers declared within a module.

        """
        return sorted({ m[1] for m in inspect.getmembers(module) if inspect.isfunction(m[1]) })


    def get_valid_types(self):
        """Return set of valid types.

        """
        return [self.get_type_name(defn[1], defn[0]) for defn in self.types]



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.schema_validation.type_validator
   :platform: Unix, Windows
   :synopsis: Validates ontology type definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from . import class_validator
from . import enum_validator
from pyesdoc.mp.ontologies.schemas import utils



# Whitelist of valid types.
_TYPE_WHITELIST = {'class', 'enum'}

# Whitelist of injected types.
_INJECTED_TYPES_WHITELIST = {'shared.doc_reference',}


def _validate_base_class_references(ctx):
    """Validates base class references.

    """
    valid_classes = [ctx.get_type_name(factory, module)
                     for module, factory, cls in ctx.classes]

    for module, factory, cls in ctx.classes:
        if 'base' in cls and cls['base'] is not None and cls['base'] not in valid_classes:
            err = 'Invalid class: {0} --> base class "{1}" is unrecognized'
            err = err.format(ctx.get_name(factory, module), cls['base'])
            ctx.set_error(err)


def _validate_class_property_type_references(ctx):
    """Validates base class references.

    """
    valid_types = ctx.get_valid_types()
    for module, factory, cls in ctx.classes:
        for name, typeof in ctx.get_class_property_types(cls):
            typeof, qualifier = utils.parse_type(typeof)
            if typeof in _INJECTED_TYPES_WHITELIST:
                continue
            if typeof not in valid_types:
                err = 'Invalid class property: {0}.[{1}] --> '
                err += 'type reference "{2}" is unrecognized'
                err = err.format(ctx.get_name(factory, module), name, typeof)
                ctx.set_error(err)
            if qualifier and qualifier not in valid_types:
                print name, typeof, qualifier
                # for t in valid_types:
                #     print t
                err = 'Invalid linked_to qualifier: {0}.[{1}] --> '
                err += 'type reference "{2}" is unrecognized'
                err = err.format(ctx.get_name(factory, module), name, qualifier)
                ctx.set_error(err)


def _validate_type(ctx, module, factory, type_):
    """Asserts package types.

    """
    if not factory.__doc__ or not factory.__doc__.strip():
        err = 'Invalid type: {} --> must specify a doc string'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)

    if 'type' not in type_:
        err = 'Invalid type: {} --> must specify a type attribute'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    if type_['type'] not in _TYPE_WHITELIST:
        err = 'Invalid type: {0} --> type attribute must be in {1}'
        err = err.format(ctx.get_name(factory, module), _TYPE_WHITELIST)
        ctx.set_error(err)
        return

    if type_['type'] == 'class':
        class_validator.validate(ctx, module, factory, type_)

    if type_['type'] == 'enum':
        enum_validator.validate(ctx, module, factory, type_)


def validate(ctx):
    """Asserts package types.

    """
    for module, factory, type_ in ctx.types:
        _validate_type(ctx, module, factory, type_)
    _validate_base_class_references(ctx)
    _validate_class_property_type_references(ctx)



# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.archive.parsing.__init__.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes a document parser that 'corrects' either content
   or decoded documents prior to further processing.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.archive.parsing import content_parser
from pyesdoc.archive.parsing import doc_parser



# Set of parsers mapped by type.
_PARSERS = {
    'content': content_parser,
    'document': doc_parser
}


def parse(target, encoding, project, source, target_type='document'):
    """Performs a parse over the target.

    :param target: Object to be parsed.
    :param encoding: Original encoding of target.
    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param target_type: Type of object to be parsed.

    :returns: Parsed target.
    :rtype: object

    """
    if target_type not in _PARSERS:
        raise KeyError("Parsing target type not supported.")

    return _PARSERS[target_type].parse(encoding, project, source, target)



# -*- coding: utf-8 -*-

"""
.. module:: factory.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: CIM v1 HTML field set factory.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc._codecs.html.fieldsets import cim_1
from pyesdoc._codecs.html.fieldsets import cim_2



# Fieldsets by ontology type.
_FIELDSETS = dict(cim_1.FIELDSETS.items() + cim_2.FIELDSETS.items())


def create(fieldset):
    """Creates a templating fieldset.

    :param fieldset: Fieldset identifier.
    :ptype fieldset: tuple | str

    :returns: A templating fieldset.
    :rtype: list

    """
    if isinstance(fieldset, tuple):
        key, obj = fieldset
        return _FIELDSETS[key](obj)
    elif isinstance(fieldset, str):
        return _FIELDSETS[fieldset]
    else:
        return fieldset or []
